@@ -22,9 +22,12 @@
 #include <stdlib.h>
 #include <stdint.h>
 #include <string.h>
+#include <errno.h>
 
 #define grub_dprintf(cond, args...) printf ( args )
 #define grub_printf printf
+#define grub_util_fopen fopen
+#define grub_memcmp memcmp
 typedef uint64_t grub_uint64_t;
 typedef uint32_t grub_uint32_t;
 typedef uint16_t grub_uint16_t;
@@ -41,6 +44,7 @@ typedef uint8_t grub_uint8_t;
 #endif
 
 #ifndef GRUB_DSDT_TEST
+#include <grub/mm.h>
 #include <grub/misc.h>
 #include <grub/time.h>
 #include <grub/cpu/io.h>
@@ -133,6 +137,49 @@ skip_data_ref_object (const grub_uint8_t *ptr, const grub_uint8_t *end)
     }
 }
 
+static inline grub_uint32_t
+skip_term (const grub_uint8_t *ptr, const grub_uint8_t *end)
+{
+  grub_uint32_t add;
+  const grub_uint8_t *ptr0 = ptr;
+
+  switch(*ptr)
+  {
+    case GRUB_ACPI_OPCODE_ADD:
+    case GRUB_ACPI_OPCODE_AND:
+    case GRUB_ACPI_OPCODE_CONCAT:
+    case GRUB_ACPI_OPCODE_CONCATRES:
+    case GRUB_ACPI_OPCODE_DIVIDE:
+    case GRUB_ACPI_OPCODE_INDEX:
+    case GRUB_ACPI_OPCODE_LSHIFT:
+    case GRUB_ACPI_OPCODE_MOD:
+    case GRUB_ACPI_OPCODE_MULTIPLY:
+    case GRUB_ACPI_OPCODE_NAND:
+    case GRUB_ACPI_OPCODE_NOR:
+    case GRUB_ACPI_OPCODE_OR:
+    case GRUB_ACPI_OPCODE_RSHIFT:
+    case GRUB_ACPI_OPCODE_SUBTRACT:
+    case GRUB_ACPI_OPCODE_TOSTRING:
+    case GRUB_ACPI_OPCODE_XOR:
+      /*
+       * Parameters for these opcodes: TermArg, TermArg Target, see ACPI
+       * spec r5.0, page 828f.
+       */
+      ptr++;
+      ptr += add = skip_term (ptr, end);
+      if (!add)
+        return 0;
+      ptr += add = skip_term (ptr, end);
+      if (!add)
+        return 0;
+      ptr += skip_name_string (ptr, end);
+      break;
+    default:
+      return skip_data_ref_object (ptr, end);
+  }
+  return ptr - ptr0;
+}
+
 static inline grub_uint32_t
 skip_ext_op (const grub_uint8_t *ptr, const grub_uint8_t *end)
 {
@@ -146,19 +193,28 @@ skip_ext_op (const grub_uint8_t *ptr, const grub_uint8_t *end)
       ptr += skip_name_string (ptr, end);
       ptr++;
       break;
+    case GRUB_ACPI_EXTOPCODE_EVENT_OP:
+      ptr++;
+      ptr += skip_name_string (ptr, end);
+      break;
     case GRUB_ACPI_EXTOPCODE_OPERATION_REGION:
       ptr++;
       ptr += skip_name_string (ptr, end);
       ptr++;
-      ptr += add = skip_data_ref_object (ptr, end);
+      ptr += add = skip_term (ptr, end);
       if (!add)
 	return 0;
-      ptr += add = skip_data_ref_object (ptr, end);
+      ptr += add = skip_term (ptr, end);
       if (!add)
 	return 0;
       break;
     case GRUB_ACPI_EXTOPCODE_FIELD_OP:
+    case GRUB_ACPI_EXTOPCODE_DEVICE_OP:
+    case GRUB_ACPI_EXTOPCODE_PROCESSOR_OP:
+    case GRUB_ACPI_EXTOPCODE_POWER_RES_OP:
+    case GRUB_ACPI_EXTOPCODE_THERMAL_ZONE_OP:
     case GRUB_ACPI_EXTOPCODE_INDEX_FIELD_OP:
+    case GRUB_ACPI_EXTOPCODE_BANK_FIELD_OP:
       ptr++;
       ptr += decode_length (ptr, 0);
       break;
@@ -169,13 +225,15 @@ skip_ext_op (const grub_uint8_t *ptr, const grub_uint8_t *end)
   return ptr - ptr0;
 }
 
+
 static int
-get_sleep_type (grub_uint8_t *table, grub_uint8_t *end)
+get_sleep_type (grub_uint8_t *table, grub_uint8_t *ptr, grub_uint8_t *end,
+		grub_uint8_t *scope, int scope_len)
 {
-  grub_uint8_t *ptr, *prev = table;
-  int sleep_type = -1;
+  grub_uint8_t *prev = table;
   
-  ptr = table + sizeof (struct grub_acpi_table_header);
+  if (!ptr)
+    ptr = table + sizeof (struct grub_acpi_table_header);
   while (ptr < end && prev < ptr)
     {
       int add;
@@ -190,6 +248,7 @@ get_sleep_type (grub_uint8_t *table, grub_uint8_t *end)
 	  if (!add)
 	    return -1;
 	  break;
+	case GRUB_ACPI_OPCODE_CREATE_DWORD_FIELD:
 	case GRUB_ACPI_OPCODE_CREATE_WORD_FIELD:
 	case GRUB_ACPI_OPCODE_CREATE_BYTE_FIELD:
 	  {
@@ -202,7 +261,8 @@ get_sleep_type (grub_uint8_t *table, grub_uint8_t *end)
 	  }
 	case GRUB_ACPI_OPCODE_NAME:
 	  ptr++;
-	  if (memcmp (ptr, "_S5_", 4) == 0 || memcmp (ptr, "\\_S5_", 4) == 0)
+	  if ((!scope || grub_memcmp (scope, "\\", scope_len) == 0) &&
+	      (grub_memcmp (ptr, "_S5_", 4) == 0 || grub_memcmp (ptr, "\\_S5_", 4) == 0))
 	    {
 	      int ll;
 	      grub_uint8_t *ptr2 = ptr;
@@ -220,14 +280,11 @@ get_sleep_type (grub_uint8_t *table, grub_uint8_t *end)
 	      switch (*ptr2)
 		{
 		case GRUB_ACPI_OPCODE_ZERO:
-		  sleep_type = 0;
-		  break;
+		  return 0;
 		case GRUB_ACPI_OPCODE_ONE:
-		  sleep_type = 1;
-		  break;
+		  return 1;
 		case GRUB_ACPI_OPCODE_BYTE_CONST:
-		  sleep_type = ptr2[1];
-		  break;
+		  return ptr2[1];
 		default:
 		  grub_printf ("Unknown data type in _S5: 0x%x\n", *ptr2);
 		  return -1;
@@ -240,7 +297,37 @@ get_sleep_type (grub_uint8_t *table, grub_uint8_t *end)
 	  if (!add)
 	    return -1;
 	  break;
+	case GRUB_ACPI_OPCODE_ALIAS:
+	  ptr++;
+	  /* We need to skip two name strings */
+	  ptr += add = skip_name_string (ptr, end);
+	  if (!add)
+	    return -1;
+	  ptr += add = skip_name_string (ptr, end);
+	  if (!add)
+	    return -1;
+	  break;
+
 	case GRUB_ACPI_OPCODE_SCOPE:
+	  {
+	    int scope_sleep_type;
+	    int ll;
+	    grub_uint8_t *name;
+	    int name_len;
+
+	    ptr++;
+	    add = decode_length (ptr, &ll);
+	    name = ptr + ll;
+	    name_len = skip_name_string (name, ptr + add);
+	    if (!name_len)
+	      return -1;
+	    scope_sleep_type = get_sleep_type (table, name + name_len,
+					       ptr + add, name, name_len);
+	    if (scope_sleep_type != -2)
+	      return scope_sleep_type;
+	    ptr += add;
+	    break;
+	  }
 	case GRUB_ACPI_OPCODE_IF:
 	case GRUB_ACPI_OPCODE_METHOD:
 	  {
@@ -254,8 +341,7 @@ get_sleep_type (grub_uint8_t *table, grub_uint8_t *end)
 	}
     }
 
-  grub_dprintf ("acpi", "TYP = %d\n", sleep_type);
-  return sleep_type;
+  return -2;
 }
 
 #ifdef GRUB_DSDT_TEST
@@ -267,7 +353,7 @@ main (int argc, char **argv)
   unsigned char *buf;
   if (argc < 2)
     printf ("Usage: %s FILE\n", argv[0]);
-  f = fopen (argv[1], "rb");
+  f = grub_util_fopen (argv[1], "rb");
   if (!f)
     {
       printf ("Couldn't open file\n");
@@ -279,19 +365,19 @@ main (int argc, char **argv)
   buf = malloc (len);
   if (!buf)
     {
-      printf ("Couldn't malloc buffer\n");
+      printf (_("error: %s.\n"), _("out of memory"));
       fclose (f);
       return 2;
     }
   if (fread (buf, 1, len, f) != len)
     {
-      printf ("Read failed\n");
+      printf (_("cannot read `%s': %s"), argv[1], strerror (errno));
       free (buf);
       fclose (f);
       return 2;
     }
 
-  printf ("Sleep type = %d\n", get_sleep_type (buf, buf + len));
+  printf ("Sleep type = %d\n", get_sleep_type (buf, NULL, buf + len, NULL, 0));
   free (buf);
   fclose (f);
   return 0;
@@ -304,8 +390,10 @@ grub_acpi_halt (void)
 {
   struct grub_acpi_rsdp_v20 *rsdp2;
   struct grub_acpi_rsdp_v10 *rsdp1;
-      struct grub_acpi_table_header *rsdt;
-      grub_uint32_t *entry_ptr;
+  struct grub_acpi_table_header *rsdt;
+  grub_uint32_t *entry_ptr;
+  grub_uint32_t port = 0;
+  int sleep_type = -1;
 
   rsdp2 = grub_acpi_get_rsdpv2 ();
   if (rsdp2)
@@ -324,35 +412,40 @@ grub_acpi_halt (void)
     {
       if (grub_memcmp ((void *) (grub_addr_t) *entry_ptr, "FACP", 4) == 0)
 	{
-	  grub_uint32_t port;
 	  struct grub_acpi_fadt *fadt
 	    = ((struct grub_acpi_fadt *) (grub_addr_t) *entry_ptr);
 	  struct grub_acpi_table_header *dsdt
 	    = (struct grub_acpi_table_header *) (grub_addr_t) fadt->dsdt_addr;
-	  int sleep_type = -1;
+	  grub_uint8_t *buf = (grub_uint8_t *) dsdt;
 
 	  port = fadt->pm1a;
 
 	  grub_dprintf ("acpi", "PM1a port=%x\n", port);
 
 	  if (grub_memcmp (dsdt->signature, "DSDT",
-			   sizeof (dsdt->signature)) != 0)
-	    break;
+			   sizeof (dsdt->signature)) == 0
+	      && sleep_type < 0)
+	    sleep_type = get_sleep_type (buf, NULL, buf + dsdt->length,
+					 NULL, 0);
+	}
+      else if (grub_memcmp ((void *) (grub_addr_t) *entry_ptr, "SSDT", 4) == 0
+	       && sleep_type < 0)
+	{
+	  struct grub_acpi_table_header *ssdt
+	    = (struct grub_acpi_table_header *) (grub_addr_t) *entry_ptr;
+	  grub_uint8_t *buf = (grub_uint8_t *) ssdt;
 
-	  sleep_type = get_sleep_type ((grub_uint8_t *) dsdt,
-				       (grub_uint8_t *) dsdt + dsdt->length);
+	  grub_dprintf ("acpi", "SSDT = %p\n", ssdt);
 
-	  if (sleep_type < 0 || sleep_type >= 8)
-	    break;
-
-	  grub_dprintf ("acpi", "SLP_TYP = %d, port = 0x%x\n",
-			sleep_type, port);
-
-	  grub_outw (GRUB_ACPI_SLP_EN
-	  	     | (sleep_type << GRUB_ACPI_SLP_TYP_OFFSET), port & 0xffff);
+	  sleep_type = get_sleep_type (buf, NULL, buf + ssdt->length, NULL, 0);
 	}
     }
 
+  grub_dprintf ("acpi", "SLP_TYP = %d, port = 0x%x\n", sleep_type, port);
+  if (port && sleep_type >= 0 && sleep_type < 8)
+    grub_outw (GRUB_ACPI_SLP_EN | (sleep_type << GRUB_ACPI_SLP_TYP_OFFSET),
+	       port & 0xffff);
+
   grub_millisleep (1500);
 
   /* TRANSLATORS: It's computer shutdown using ACPI, not disabling ACPI.  */
