@@ -35,6 +35,15 @@
 
 GRUB_MOD_LICENSE ("GPLv3+");
 
+/* Helper for legacy_file.  */
+static grub_err_t
+legacy_file_getline (char **line, int cont __attribute__ ((unused)),
+		     void *data __attribute__ ((unused)))
+{
+  *line = 0;
+  return GRUB_ERR_NONE;
+}
+
 static grub_err_t
 legacy_file (const char *filename)
 {
@@ -43,27 +52,25 @@ legacy_file (const char *filename)
   grub_menu_t menu;
   char *suffix = grub_strdup ("");
 
-  auto grub_err_t getline (char **line, int cont);
-  grub_err_t getline (char **line, 
-		      int cont __attribute__ ((unused)))
-  {
-    *line = 0;
-    return GRUB_ERR_NONE;
-  }
-
   if (!suffix)
     return grub_errno;
 
   file = grub_file_open (filename);
   if (! file)
-    return grub_errno;
+    {
+      grub_free (suffix);
+      return grub_errno;
+    }
 
   menu = grub_env_get_menu ();
   if (! menu)
     {
       menu = grub_zalloc (sizeof (*menu));
       if (! menu)
-	return grub_errno;
+	{
+	  grub_free (suffix);
+	  return grub_errno;
+	}
 
       grub_env_set_menu (menu);
     }
@@ -76,6 +83,7 @@ legacy_file (const char *filename)
       if (!buf && grub_errno)
 	{
 	  grub_file_close (file);
+	  grub_free (suffix);
 	  return grub_errno;
 	}
 
@@ -134,7 +142,7 @@ legacy_file (const char *filename)
 
       if (parsed && !entryname)
 	{
-	  grub_normal_parse_line (parsed, getline);
+	  grub_normal_parse_line (parsed, legacy_file_getline, NULL);
 	  grub_print_error ();
 	  grub_free (parsed);
 	  parsed = NULL;
@@ -172,6 +180,8 @@ legacy_file (const char *filename)
       if (!args)
 	{
 	  grub_file_close (file);
+	  grub_free (suffix);
+	  grub_free (entrysrc);
 	  return grub_errno;
 	}
       args[0] = entryname;
@@ -180,7 +190,7 @@ legacy_file (const char *filename)
       grub_free (args);
     }
 
-  grub_normal_parse_line (suffix, getline);
+  grub_normal_parse_line (suffix, legacy_file_getline, NULL);
   grub_print_error ();
   grub_free (suffix);
   grub_free (entrysrc);
@@ -243,6 +253,7 @@ grub_cmd_legacy_kernel (struct grub_command *mycmd __attribute__ ((unused)),
   struct grub_command *cmd;
   char **cutargs;
   int cutargc;
+  grub_err_t err = GRUB_ERR_NONE;
   
   for (i = 0; i < 2; i++)
     {
@@ -304,6 +315,8 @@ grub_cmd_legacy_kernel (struct grub_command *mycmd __attribute__ ((unused)),
     return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("filename expected"));
 
   cutargs = grub_malloc (sizeof (cutargs[0]) * (argc - 1));
+  if (!cutargs)
+    return grub_errno;
   cutargc = argc - 1;
   grub_memcpy (cutargs + 1, args + 2, sizeof (cutargs[0]) * (argc - 2));
   cutargs[0] = args[0];
@@ -313,13 +326,17 @@ grub_cmd_legacy_kernel (struct grub_command *mycmd __attribute__ ((unused)),
       /* First try Linux.  */
       if (kernel_type == GUESS_IT || kernel_type == LINUX)
 	{
+#ifdef GRUB_MACHINE_PCBIOS
 	  cmd = grub_command_find ("linux16");
+#else
+	  cmd = grub_command_find ("linux");
+#endif
 	  if (cmd)
 	    {
 	      if (!(cmd->func) (cmd, cutargc, cutargs))
 		{
 		  kernel_type = LINUX;
-		  return GRUB_ERR_NONE;
+		  goto out;
 		}
 	    }
 	  grub_errno = GRUB_ERR_NONE;
@@ -334,7 +351,7 @@ grub_cmd_legacy_kernel (struct grub_command *mycmd __attribute__ ((unused)),
 	      if (!(cmd->func) (cmd, argc, args))
 		{
 		  kernel_type = MULTIBOOT;
-		  return GRUB_ERR_NONE;
+		  goto out;
 		}
 	    }
 	  grub_errno = GRUB_ERR_NONE;
@@ -357,7 +374,7 @@ grub_cmd_legacy_kernel (struct grub_command *mycmd __attribute__ ((unused)),
 	  dev = grub_device_open (0);
 	  if (dev && dev->disk
 	      && dev->disk->dev->id == GRUB_DISK_DEVICE_BIOSDISK_ID
-	      && dev->disk->dev->id >= 0x80 && dev->disk->dev->id <= 0x90)
+	      && dev->disk->id >= 0x80 && dev->disk->id <= 0x90)
 	    {
 	      struct grub_partition *part = dev->disk->partition;
 	      bsd_device = dev->disk->id - 0x80 - hdbias;
@@ -371,6 +388,8 @@ grub_cmd_legacy_kernel (struct grub_command *mycmd __attribute__ ((unused)),
 	      if (part && grub_strcmp (part->partmap->name, "msdos") == 0)
 		bsd_slice = part->number;
 	    }
+	  if (dev)
+	    grub_device_close (dev);
 	}
 	
 	/* k*BSD didn't really work well with grub-legacy.  */
@@ -397,7 +416,7 @@ grub_cmd_legacy_kernel (struct grub_command *mycmd __attribute__ ((unused)),
 		if (!(cmd->func) (cmd, cutargc, cutargs))
 		  {
 		    kernel_type = KFREEBSD;
-		    return GRUB_ERR_NONE;
+		    goto out;
 		  }
 	      }
 	    grub_errno = GRUB_ERR_NONE;
@@ -406,6 +425,8 @@ grub_cmd_legacy_kernel (struct grub_command *mycmd __attribute__ ((unused)),
 	  char **bsdargs;
 	  int bsdargc;
 	  char bsddevname[sizeof ("wdXXXXXXXXXXXXY")];
+	  int found = 0;
+
 	  if (bsd_device == -1)
 	    {
 	      bsdargs = cutargs;
@@ -416,6 +437,11 @@ grub_cmd_legacy_kernel (struct grub_command *mycmd __attribute__ ((unused)),
 	      char rbuf[3] = "-r";
 	      bsdargc = cutargc + 2;
 	      bsdargs = grub_malloc (sizeof (bsdargs[0]) * bsdargc);
+	      if (!bsdargs)
+		{
+		  err = grub_errno;
+		  goto out;
+		}
 	      grub_memcpy (bsdargs, args, argc * sizeof (bsdargs[0]));
 	      bsdargs[argc] = rbuf;
 	      bsdargs[argc + 1] = bsddevname;
@@ -431,7 +457,8 @@ grub_cmd_legacy_kernel (struct grub_command *mycmd __attribute__ ((unused)),
 		  if (!(cmd->func) (cmd, bsdargc, bsdargs))
 		    {
 		      kernel_type = KNETBSD;
-		      return GRUB_ERR_NONE;
+		      found = 1;
+		      goto free_bsdargs;
 		    }
 		}
 	      grub_errno = GRUB_ERR_NONE;
@@ -444,20 +471,28 @@ grub_cmd_legacy_kernel (struct grub_command *mycmd __attribute__ ((unused)),
 		  if (!(cmd->func) (cmd, bsdargc, bsdargs))
 		    {
 		      kernel_type = KOPENBSD;
-		      return GRUB_ERR_NONE;
+		      found = 1;
+		      goto free_bsdargs;
 		    }
 		}
 	      grub_errno = GRUB_ERR_NONE;
 	    }
+
+free_bsdargs:
 	  if (bsdargs != cutargs)
 	    grub_free (bsdargs);
+	  if (found)
+	    goto out;
 	}
       }
     }
   while (0);
 
-  return grub_error (GRUB_ERR_BAD_OS, "couldn't load file %s",
-		     args[0]);
+  err = grub_error (GRUB_ERR_BAD_OS, "couldn't load file %s",
+		    args[0]);
+out:
+  grub_free (cutargs);
+  return err;
 }
 
 static grub_err_t
@@ -468,12 +503,21 @@ grub_cmd_legacy_initrd (struct grub_command *mycmd __attribute__ ((unused)),
 
   if (kernel_type == LINUX)
     {
+#ifdef GRUB_MACHINE_PCBIOS
       cmd = grub_command_find ("initrd16");
+#else
+      cmd = grub_command_find ("initrd");
+#endif
       if (!cmd)
 	return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("can't find command `%s'"),
-			   "initrd16");
+#ifdef GRUB_MACHINE_PCBIOS
+			   "initrd16"
+#else
+			   "initrd"
+#endif
+			   );
 
-      return cmd->func (cmd, argc, args);
+      return cmd->func (cmd, argc ? 1 : 0, args);
     }
   if (kernel_type == MULTIBOOT)
     {
@@ -509,15 +553,17 @@ grub_cmd_legacy_initrdnounzip (struct grub_command *mycmd __attribute__ ((unused
       char **newargs;
       grub_err_t err;
       char nounzipbuf[10] = "--nounzip";
+
+      cmd = grub_command_find ("module");
+      if (!cmd)
+	return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("can't find command `%s'"),
+			   "module");
+
       newargs = grub_malloc ((argc + 1) * sizeof (newargs[0]));
       if (!newargs)
 	return grub_errno;
       grub_memcpy (newargs + 1, args, argc * sizeof (newargs[0]));
       newargs[0] = nounzipbuf;
-      cmd = grub_command_find ("module");
-      if (!cmd)
-	return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("can't find command `%s'"),
-			   "module");
 
       err = cmd->func (cmd, argc + 1, newargs);
       grub_free (newargs);
@@ -545,8 +591,6 @@ struct legacy_md5_password
   grub_uint8_t hash[MD5_HASHLEN];
 };
 
-#pragma GCC diagnostic ignored "-Wunsafe-loop-optimizations"
-
 static int
 check_password_md5_real (const char *entered,
 			 struct legacy_md5_password *pw)
@@ -554,8 +598,13 @@ check_password_md5_real (const char *entered,
   grub_size_t enteredlen = grub_strlen (entered);
   unsigned char alt_result[MD5_HASHLEN];
   unsigned char *digest;
-  grub_uint8_t ctx[GRUB_MD_MD5->contextsize];
+  grub_uint8_t *ctx;
   grub_size_t i;
+  int ret;
+
+  ctx = grub_zalloc (GRUB_MD_MD5->contextsize);
+  if (!ctx)
+    return 0;
 
   GRUB_MD_MD5->init (ctx);
   GRUB_MD_MD5->write (ctx, entered, enteredlen);
@@ -563,7 +612,7 @@ check_password_md5_real (const char *entered,
   GRUB_MD_MD5->write (ctx, entered, enteredlen);
   digest = GRUB_MD_MD5->read (ctx);
   GRUB_MD_MD5->final (ctx);
-  memcpy (alt_result, digest, MD5_HASHLEN);
+  grub_memcpy (alt_result, digest, MD5_HASHLEN);
   
   GRUB_MD_MD5->init (ctx);
   GRUB_MD_MD5->write (ctx, entered, enteredlen);
@@ -579,7 +628,7 @@ check_password_md5_real (const char *entered,
 
   for (i = 0; i < 1000; i++)
     {
-      memcpy (alt_result, digest, 16);
+      grub_memcpy (alt_result, digest, 16);
 
       GRUB_MD_MD5->init (ctx);
       if ((i & 1) != 0)
@@ -601,7 +650,9 @@ check_password_md5_real (const char *entered,
       GRUB_MD_MD5->final (ctx);
     }
 
-  return (grub_crypto_memcmp (digest, pw->hash, MD5_HASHLEN) == 0);
+  ret = (grub_crypto_memcmp (digest, pw->hash, MD5_HASHLEN) == 0);
+  grub_free (ctx);
+  return ret;
 }
 
 static grub_err_t
@@ -724,18 +775,12 @@ grub_cmd_legacy_password (struct grub_command *mycmd __attribute__ ((unused)),
 					      NULL);
 }
 
-static grub_err_t
-grub_cmd_legacy_check_password (struct grub_command *mycmd __attribute__ ((unused)),
-				int argc, char **args)
+int
+grub_legacy_check_md5_password (int argc, char **args,
+				char *entered)
 {
   struct legacy_md5_password *pw = NULL;
-  char entered[GRUB_AUTH_MAX_PASSLEN];
-
-  if (argc == 0)
-    return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("one argument expected"));
-  grub_puts_ (N_("Enter password: "));
-  if (!grub_password_get (entered, GRUB_AUTH_MAX_PASSLEN))
-    return GRUB_ACCESS_DENIED;
+  int ret;
 
   if (args[0][0] != '-' || args[0][1] != '-')
     {
@@ -744,17 +789,33 @@ grub_cmd_legacy_check_password (struct grub_command *mycmd __attribute__ ((unuse
       grub_memset (correct, 0, sizeof (correct));
       grub_strncpy (correct, args[0], sizeof (correct));
 
-      if (grub_crypto_memcmp (entered, correct, GRUB_AUTH_MAX_PASSLEN) != 0)
-	return GRUB_ACCESS_DENIED;
-      return GRUB_ERR_NONE;
+      return grub_crypto_memcmp (entered, correct, GRUB_AUTH_MAX_PASSLEN) == 0;
     }
 
   pw = parse_legacy_md5 (argc, args);
 
   if (!pw)
+    return 0;
+
+  ret = check_password_md5_real (entered, pw);
+  grub_free (pw);
+  return ret;
+}
+
+static grub_err_t
+grub_cmd_legacy_check_password (struct grub_command *mycmd __attribute__ ((unused)),
+				int argc, char **args)
+{
+  char entered[GRUB_AUTH_MAX_PASSLEN];
+
+  if (argc == 0)
+    return grub_error (GRUB_ERR_BAD_ARGUMENT, N_("one argument expected"));
+  grub_puts_ (N_("Enter password: "));
+  if (!grub_password_get (entered, GRUB_AUTH_MAX_PASSLEN))
     return GRUB_ACCESS_DENIED;
 
-  if (!check_password_md5_real (entered, pw))
+  if (!grub_legacy_check_md5_password (argc, args,
+				       entered))
     return GRUB_ACCESS_DENIED;
 
   return GRUB_ERR_NONE;
