@@ -363,8 +363,6 @@ static ush mask_bits[] =
   0x01ff, 0x03ff, 0x07ff, 0x0fff, 0x1fff, 0x3fff, 0x7fff, 0xffff
 };
 
-#pragma GCC diagnostic ignored "-Wunsafe-loop-optimizations"
-
 #define NEEDBITS(n) do {while(k<(n)){b|=((ulg)get_byte(gzio))<<k;k+=8;}} while (0)
 #define DUMPBITS(n) do {b>>=(n);k-=(n);} while (0)
 
@@ -544,7 +542,7 @@ huft_build (unsigned *b,	/* code lengths in bits (all assumed <= BMAX) */
 	      z = 1 << j;	/* table entries for j-bit table */
 
 	      /* allocate and link in new table */
-	      q = (struct huft *) grub_malloc ((z + 1) * sizeof (struct huft));
+	      q = (struct huft *) grub_zalloc ((z + 1) * sizeof (struct huft));
 	      if (! q)
 		{
 		  if (h)
@@ -1095,7 +1093,7 @@ inflate_window (grub_gzio_t gzio)
 	}
     }
 
-  gzio->saved_offset += WSIZE;
+  gzio->saved_offset += gzio->wp;
 
   /* XXX do CRC calculation here! */
 }
@@ -1118,6 +1116,8 @@ initialize_tables (grub_gzio_t gzio)
   /* Reset memory allocation stuff.  */
   huft_free (gzio->tl);
   huft_free (gzio->td);
+  gzio->tl = NULL;
+  gzio->td = NULL;
 }
 
 
@@ -1161,6 +1161,19 @@ grub_gzio_open (grub_file_t io, const char *name __attribute__ ((unused)))
   return file;
 }
 
+static grub_uint8_t
+mod_31 (grub_uint16_t v)
+{
+  /* At most 2 iterations for any number that
+     we can get here.
+     In any case faster than real division.  */
+  while (v > 0x1f)
+    v = (v & 0x1f) + (v >> 5);
+  if (v == 0x1f)
+    return 0;
+  return v;
+}
+
 static int
 test_zlib_header (grub_gzio_t gzio)
 {
@@ -1178,7 +1191,10 @@ test_zlib_header (grub_gzio_t gzio)
       return 0;
     }
 
-  if ((cmf * 256 + flg) % 31)
+  /* Usually it would be: (cmf * 256 + flg) % 31 != 0.  */
+  /* But 256 == 8 (31).  */
+  /* By multiplying by 4 and using 32 == 1 (31). We get our formula.  */
+  if (mod_31 (cmf + flg * 4) != 0)
     {
       grub_error (GRUB_ERR_BAD_COMPRESSED_DATA, N_("unsupported gzip format"));
       return 0;
@@ -1219,7 +1235,14 @@ grub_gzio_read_real (grub_gzio_t gzio, grub_off_t offset,
       register char *srcaddr;
 
       while (offset >= gzio->saved_offset)
-	inflate_window (gzio);
+	{
+	  inflate_window (gzio);
+	  if (gzio->wp == 0)
+	    goto out;
+	}
+
+      if (gzio->wp == 0)
+	goto out;
 
       srcaddr = (char *) ((offset & (WSIZE - 1)) + gzio->slide);
       size = gzio->saved_offset - offset;
@@ -1234,6 +1257,7 @@ grub_gzio_read_real (grub_gzio_t gzio, grub_off_t offset,
       offset += size;
     }
 
+ out:
   if (grub_errno != GRUB_ERR_NONE)
     ret = -1;
 
@@ -1243,7 +1267,15 @@ grub_gzio_read_real (grub_gzio_t gzio, grub_off_t offset,
 static grub_ssize_t
 grub_gzio_read (grub_file_t file, char *buf, grub_size_t len)
 {
-  return grub_gzio_read_real (file->data, file->offset, buf, len);
+  grub_ssize_t ret;
+  ret = grub_gzio_read_real (file->data, file->offset, buf, len);
+
+  if (!grub_errno && ret != (grub_ssize_t) len)
+    {
+      grub_error (GRUB_ERR_BAD_COMPRESSED_DATA, "premature end of compressed");
+      ret = -1;
+    }
+  return ret;
 }
 
 /* Release everything, including the underlying file object.  */
@@ -1259,6 +1291,7 @@ grub_gzio_close (grub_file_t file)
 
   /* No need to close the same device twice.  */
   file->device = 0;
+  file->name = 0;
 
   return grub_errno;
 }
@@ -1290,6 +1323,28 @@ grub_zlib_decompress (char *inbuf, grub_size_t insize, grub_off_t off,
   return ret;
 }
 
+grub_ssize_t
+grub_deflate_decompress (char *inbuf, grub_size_t insize, grub_off_t off,
+			 char *outbuf, grub_size_t outsize)
+{
+  grub_gzio_t gzio = 0;
+  grub_ssize_t ret;
+
+  gzio = grub_zalloc (sizeof (*gzio));
+  if (! gzio)
+    return -1;
+  gzio->mem_input = (grub_uint8_t *) inbuf;
+  gzio->mem_input_size = insize;
+  gzio->mem_input_off = 0;
+
+  initialize_tables (gzio);
+
+  ret = grub_gzio_read_real (gzio, off, outbuf, outsize);
+  grub_free (gzio);
+
+  return ret;
+}
+
 
 
 static struct grub_fs grub_gzio_fs =
