@@ -20,18 +20,12 @@
 #include <config.h>
 
 #include <errno.h>
-#include <error.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <stdint.h>
 #include <string.h>
 #include <sys/time.h>
 #include <sys/types.h>
-#include <sys/stat.h>
-#include <unistd.h>
-#ifdef HAVE_LIMITS_H
-#include <limits.h>
-#endif
 
 #include <grub/mm.h>
 #include <grub/err.h>
@@ -42,27 +36,6 @@
 #include <grub/time.h>
 #include <grub/emu/misc.h>
 
-#ifdef HAVE_DEVICE_MAPPER
-# include <libdevmapper.h>
-#endif
-
-#ifdef HAVE_SYS_PARAM_H
-# include <sys/param.h>
-#endif
-
-#ifdef HAVE_SYS_MOUNT_H
-# include <sys/mount.h>
-#endif
-
-#ifdef HAVE_SYS_MNTTAB_H
-# include <stdio.h> /* Needed by sys/mnttab.h.  */
-# include <sys/mnttab.h>
-#endif
-
-#ifdef HAVE_SYS_MKDEV_H
-# include <sys/mkdev.h> /* makedev */
-#endif
-
 int verbosity;
 
 void
@@ -145,36 +118,7 @@ xstrdup (const char *str)
   return newstr;
 }
 
-#ifndef HAVE_VASPRINTF
-
-int
-vasprintf (char **buf, const char *fmt, va_list ap)
-{
-  /* Should be large enough.  */
-  *buf = xmalloc (512);
-
-  return vsnprintf (*buf, 512, fmt, ap);
-}
-
-#endif
-
-#ifndef  HAVE_ASPRINTF
-
-int
-asprintf (char **buf, const char *fmt, ...)
-{
-  int status;
-  va_list ap;
-
-  va_start (ap, fmt);
-  status = vasprintf (buf, fmt, ap);
-  va_end (ap);
-
-  return status;
-}
-
-#endif
-
+#if !defined (GRUB_MKFONT) && !defined (GRUB_BUILD)
 char *
 xasprintf (const char *fmt, ...)
 { 
@@ -182,21 +126,22 @@ xasprintf (const char *fmt, ...)
   char *result;
   
   va_start (ap, fmt);
-  if (vasprintf (&result, fmt, ap) < 0)
-    { 
-      if (errno == ENOMEM)
-        grub_util_error ("%s", _("out of memory"));
-      return NULL;
-    }
+  result = grub_xvasprintf (fmt, ap);
+  va_end (ap);
+  if (!result)
+    grub_util_error ("%s", _("out of memory"));
   
   return result;
 }
+#endif
 
+#if !defined (GRUB_MACHINE_EMU) || defined (GRUB_UTIL)
 void
 grub_exit (void)
 {
   exit (1);
 }
+#endif
 
 grub_uint64_t
 grub_get_time_ms (void)
@@ -208,67 +153,50 @@ grub_get_time_ms (void)
   return (tv.tv_sec * 1000 + tv.tv_usec / 1000);
 }
 
-grub_uint32_t
-grub_get_rtc (void)
+size_t
+grub_util_get_image_size (const char *path)
 {
-  struct timeval tv;
+  FILE *f;
+  size_t ret;
+  off_t sz;
 
-  gettimeofday (&tv, 0);
+  f = grub_util_fopen (path, "rb");
 
-  return (tv.tv_sec * GRUB_TICKS_PER_SECOND
-	  + (((tv.tv_sec % GRUB_TICKS_PER_SECOND) * 1000000 + tv.tv_usec)
-	     * GRUB_TICKS_PER_SECOND / 1000000));
-}
+  if (!f)
+    grub_util_error (_("cannot open `%s': %s"), path, strerror (errno));
+
+  fseeko (f, 0, SEEK_END);
+  
+  sz = ftello (f);
+  if (sz < 0)
+    grub_util_error (_("cannot open `%s': %s"), path, strerror (errno));
+  if (sz != (size_t) sz)
+    grub_util_error (_("file `%s' is too big"), path);
+  ret = (size_t) sz;
+
+  fclose (f);
 
-char *
-canonicalize_file_name (const char *path)
-{
-  char *ret;
-#ifdef __MINGW32__
-  ret = xmalloc (PATH_MAX);
-  if (!_fullpath (ret, path, PATH_MAX))
-    return NULL;
-#elif defined (PATH_MAX)
-  ret = xmalloc (PATH_MAX);
-  if (!realpath (path, ret))
-    return NULL;
-#else
-  ret = realpath (path, NULL);
-#endif
   return ret;
 }
 
-#ifdef HAVE_DEVICE_MAPPER
-static void device_mapper_null_log (int level __attribute__ ((unused)),
-				    const char *file __attribute__ ((unused)),
-				    int line __attribute__ ((unused)),
-				    int dm_errno __attribute__ ((unused)),
-				    const char *f __attribute__ ((unused)),
-				    ...)
+void
+grub_util_load_image (const char *path, char *buf)
 {
+  FILE *fp;
+  size_t size;
+
+  grub_util_info ("reading %s", path);
+
+  size = grub_util_get_image_size (path);
+
+  fp = grub_util_fopen (path, "rb");
+  if (! fp)
+    grub_util_error (_("cannot open `%s': %s"), path,
+		     strerror (errno));
+
+  if (fread (buf, 1, size, fp) != size)
+    grub_util_error (_("cannot read `%s': %s"), path,
+		     strerror (errno));
+
+  fclose (fp);
 }
-
-int
-grub_device_mapper_supported (void)
-{
-  static int supported = -1;
-
-  if (supported == -1)
-    {
-      struct dm_task *dmt;
-
-      /* Suppress annoying log messages.  */
-      dm_log_with_errno_init (&device_mapper_null_log);
-
-      dmt = dm_task_create (DM_DEVICE_VERSION);
-      supported = (dmt != NULL);
-      if (dmt)
-	dm_task_destroy (dmt);
-
-      /* Restore the original logger.  */
-      dm_log_with_errno_init (NULL);
-    }
-
-  return supported;
-}
-#endif /* HAVE_DEVICE_MAPPER */
