@@ -34,6 +34,10 @@
 #include <grub/net.h>
 #include <grub/offsets.h>
 #include <grub/memory.h>
+#include <grub/loader.h>
+#ifdef __i386__
+#include <grub/cpu/tsc.h>
+#endif
 #ifdef __sparc__
 #include <grub/machine/kernel.h>
 #endif
@@ -42,11 +46,19 @@
 #define HEAP_MIN_SIZE		(unsigned long) (2 * 1024 * 1024)
 
 /* The maximum heap size we're going to claim */
+#ifdef __i386__
+#define HEAP_MAX_SIZE		(unsigned long) (64 * 1024 * 1024)
+#else
 #define HEAP_MAX_SIZE		(unsigned long) (32 * 1024 * 1024)
+#endif
 
 /* If possible, we will avoid claiming heap above this address, because it
    seems to cause relocation problems with OSes that link at 4 MiB */
+#ifdef __i386__
+#define HEAP_MAX_ADDR		(unsigned long) (64 * 1024 * 1024)
+#else
 #define HEAP_MAX_ADDR		(unsigned long) (32 * 1024 * 1024)
+#endif
 
 extern char _start[];
 extern char _end[];
@@ -76,24 +88,35 @@ grub_translate_ieee1275_path (char *filepath)
     }
 }
 
-void (*grub_ieee1275_net_config) (const char *dev,
-				  char **device,
-				  char **path);
+void (*grub_ieee1275_net_config) (const char *dev, char **device, char **path,
+                                  char *bootpath);
 void
 grub_machine_get_bootlocation (char **device, char **path)
 {
-  char bootpath[64]; /* XXX check length */
+  char *bootpath;
+  grub_ssize_t bootpath_size;
   char *filename;
   char *type;
-   
-  if (grub_ieee1275_get_property (grub_ieee1275_chosen, "bootpath", &bootpath,
-				  sizeof (bootpath), 0))
+
+  if (grub_ieee1275_get_property_length (grub_ieee1275_chosen, "bootpath",
+					 &bootpath_size)
+      || bootpath_size <= 0)
     {
       /* Should never happen.  */
       grub_printf ("/chosen/bootpath property missing!\n");
       return;
     }
 
+  bootpath = (char *) grub_malloc ((grub_size_t) bootpath_size + 64);
+  if (! bootpath)
+    {
+      grub_print_error ();
+      return;
+    }
+  grub_ieee1275_get_property (grub_ieee1275_chosen, "bootpath", bootpath,
+                              (grub_size_t) bootpath_size + 1, 0);
+  bootpath[bootpath_size] = '\0';
+
   /* Transform an OF device path to a GRUB path.  */
 
   type = grub_ieee1275_get_device_type (bootpath);
@@ -110,7 +133,7 @@ grub_machine_get_bootlocation (char **device, char **path)
       *ptr = 0;
 
       if (grub_ieee1275_net_config)
-	grub_ieee1275_net_config (canon, device, path);
+	grub_ieee1275_net_config (canon, device, path, bootpath);
       grub_free (dev);
       grub_free (canon);
     }
@@ -132,6 +155,7 @@ grub_machine_get_bootlocation (char **device, char **path)
 	  *path = filename;
 	}
     }
+  grub_free (bootpath);
 }
 
 /* Claim some available memory in the first /memory node. */
@@ -143,74 +167,77 @@ grub_claim_heap (void)
 				 + GRUB_KERNEL_MACHINE_STACK_SIZE), 0x200000);
 }
 #else
+/* Helper for grub_claim_heap.  */
+static int
+heap_init (grub_uint64_t addr, grub_uint64_t len, grub_memory_type_t type,
+	   void *data)
+{
+  unsigned long *total = data;
+
+  if (type != GRUB_MEMORY_AVAILABLE)
+    return 0;
+
+  if (grub_ieee1275_test_flag (GRUB_IEEE1275_FLAG_NO_PRE1_5M_CLAIM))
+    {
+      if (addr + len <= 0x180000)
+	return 0;
+
+      if (addr < 0x180000)
+	{
+	  len = addr + len - 0x180000;
+	  addr = 0x180000;
+	}
+    }
+  len -= 1; /* Required for some firmware.  */
+
+  /* Never exceed HEAP_MAX_SIZE  */
+  if (*total + len > HEAP_MAX_SIZE)
+    len = HEAP_MAX_SIZE - *total;
+
+  /* Avoid claiming anything above HEAP_MAX_ADDR, if possible. */
+  if ((addr < HEAP_MAX_ADDR) &&				/* if it's too late, don't bother */
+      (addr + len > HEAP_MAX_ADDR) &&				/* if it wasn't available anyway, don't bother */
+      (*total + (HEAP_MAX_ADDR - addr) > HEAP_MIN_SIZE))	/* only limit ourselves when we can afford to */
+     len = HEAP_MAX_ADDR - addr;
+
+  /* In theory, firmware should already prevent this from happening by not
+     listing our own image in /memory/available.  The check below is intended
+     as a safeguard in case that doesn't happen.  However, it doesn't protect
+     us from corrupting our module area, which extends up to a
+     yet-undetermined region above _end.  */
+  if ((addr < (grub_addr_t) _end) && ((addr + len) > (grub_addr_t) _start))
+    {
+      grub_printf ("Warning: attempt to claim over our own code!\n");
+      len = 0;
+    }
+
+  if (len)
+    {
+      grub_err_t err;
+      /* Claim and use it.  */
+      err = grub_claimmap (addr, len);
+      if (err)
+	return err;
+      grub_mm_init_region ((void *) (grub_addr_t) addr, len);
+    }
+
+  *total += len;
+  if (*total >= HEAP_MAX_SIZE)
+    return 1;
+
+  return 0;
+}
+
 static void 
 grub_claim_heap (void)
 {
   unsigned long total = 0;
 
-  auto int NESTED_FUNC_ATTR heap_init (grub_uint64_t addr, grub_uint64_t len,
-				       grub_memory_type_t type);
-  int NESTED_FUNC_ATTR heap_init (grub_uint64_t addr, grub_uint64_t len,
-				  grub_memory_type_t type)
-  {
-    if (type != 1)
-      return 0;
-
-    if (grub_ieee1275_test_flag (GRUB_IEEE1275_FLAG_NO_PRE1_5M_CLAIM))
-      {
-	if (addr + len <= 0x180000)
-	  return 0;
-
-	if (addr < 0x180000)
-	  {
-	    len = addr + len - 0x180000;
-	    addr = 0x180000;
-	  }
-      }
-    len -= 1; /* Required for some firmware.  */
-
-    /* Never exceed HEAP_MAX_SIZE  */
-    if (total + len > HEAP_MAX_SIZE)
-      len = HEAP_MAX_SIZE - total;
-
-    /* Avoid claiming anything above HEAP_MAX_ADDR, if possible. */
-    if ((addr < HEAP_MAX_ADDR) &&				/* if it's too late, don't bother */
-        (addr + len > HEAP_MAX_ADDR) &&				/* if it wasn't available anyway, don't bother */
-        (total + (HEAP_MAX_ADDR - addr) > HEAP_MIN_SIZE))	/* only limit ourselves when we can afford to */
-       len = HEAP_MAX_ADDR - addr;
-
-    /* In theory, firmware should already prevent this from happening by not
-       listing our own image in /memory/available.  The check below is intended
-       as a safeguard in case that doesn't happen.  However, it doesn't protect
-       us from corrupting our module area, which extends up to a
-       yet-undetermined region above _end.  */
-    if ((addr < (grub_addr_t) _end) && ((addr + len) > (grub_addr_t) _start))
-      {
-        grub_printf ("Warning: attempt to claim over our own code!\n");
-        len = 0;
-      }
-
-    if (len)
-      {
-	grub_err_t err;
-	/* Claim and use it.  */
-	err = grub_claimmap (addr, len);
-	if (err)
-	  return err;
-	grub_mm_init_region ((void *) (grub_addr_t) addr, len);
-      }
-
-    total += len;
-    if (total >= HEAP_MAX_SIZE)
-      return 1;
-
-    return 0;
-  }
-
-  if (grub_ieee1275_test_flag (GRUB_IEEE1275_FLAG_CANNOT_INTERPRET))
-    heap_init (HEAP_MAX_ADDR - HEAP_MIN_SIZE, HEAP_MIN_SIZE, 1);
+  if (grub_ieee1275_test_flag (GRUB_IEEE1275_FLAG_FORCE_CLAIM))
+    heap_init (GRUB_IEEE1275_STATIC_HEAP_START, GRUB_IEEE1275_STATIC_HEAP_LEN,
+	       1, &total);
   else
-    grub_machine_mmap_iterate (heap_init);
+    grub_machine_mmap_iterate (heap_init, &total);
 }
 #endif
 
@@ -254,8 +281,6 @@ grub_parse_cmdline (void)
     }
 }
 
-static grub_uint64_t ieee1275_get_time_ms (void);
-
 grub_addr_t grub_modbase;
 
 void
@@ -273,18 +298,25 @@ grub_machine_init (void)
 
   grub_parse_cmdline ();
 
-  grub_install_get_time_ms (ieee1275_get_time_ms);
+#ifdef __i386__
+  grub_tsc_init ();
+#else
+  grub_install_get_time_ms (grub_rtc_get_time_ms);
+#endif
 }
 
 void
-grub_machine_fini (void)
+grub_machine_fini (int flags)
 {
-  grub_ofdisk_fini ();
-  grub_console_fini ();
+  if (flags & GRUB_LOADER_FLAG_NORETURN)
+    {
+      grub_ofdisk_fini ();
+      grub_console_fini ();
+    }
 }
 
-static grub_uint64_t
-ieee1275_get_time_ms (void)
+grub_uint64_t
+grub_rtc_get_time_ms (void)
 {
   grub_uint32_t msecs = 0;
 
