@@ -37,21 +37,6 @@ GRUB_MOD_LICENSE ("GPLv3+");
 
 char *grub_net_default_server;
 
-struct grub_net_route
-{
-  struct grub_net_route *next;
-  struct grub_net_route **prev;
-  grub_net_network_level_netaddress_t target;
-  char *name;
-  struct grub_net_network_level_protocol *prot;
-  int is_gateway;
-  union
-  {
-    struct grub_net_network_level_interface *interface;
-    grub_net_network_level_address_t gw;
-  };
-};
-
 struct grub_net_route *grub_net_routes = NULL;
 struct grub_net_network_level_interface *grub_net_network_level_interfaces = NULL;
 struct grub_net_card *grub_net_cards = NULL;
@@ -243,19 +228,64 @@ grub_net_ipv6_get_slaac (struct grub_net_card *card,
   return slaac;
 }
 
+static void
+grub_net_network_level_interface_register (struct grub_net_network_level_interface *inter);
+
+static struct grub_net_network_level_interface *
+grub_net_add_addr_real (char *name, 
+			struct grub_net_card *card,
+			const grub_net_network_level_address_t *addr,
+			const grub_net_link_level_address_t *hwaddress,
+			grub_net_interface_flags_t flags)
+{
+  struct grub_net_network_level_interface *inter;
+
+  inter = grub_zalloc (sizeof (*inter));
+  if (!inter)
+    return NULL;
+
+  inter->name = name;
+  grub_memcpy (&(inter->address), addr, sizeof (inter->address));
+  grub_memcpy (&(inter->hwaddress), hwaddress, sizeof (inter->hwaddress));
+  inter->flags = flags;
+  inter->card = card;
+  inter->dhcp_ack = NULL;
+  inter->dhcp_acklen = 0;
+
+  grub_net_network_level_interface_register (inter);
+
+  return inter;
+}
+
+struct grub_net_network_level_interface *
+grub_net_add_addr (const char *name, 
+		   struct grub_net_card *card,
+		   const grub_net_network_level_address_t *addr,
+		   const grub_net_link_level_address_t *hwaddress,
+		   grub_net_interface_flags_t flags)
+{
+  char *name_dup = grub_strdup (name);
+  struct grub_net_network_level_interface *ret;
+ 
+  if (!name_dup)
+    return NULL;
+  ret = grub_net_add_addr_real (name_dup, card, addr, hwaddress, flags);
+  if (!ret)
+    grub_free (name_dup);
+  return ret;
+}
+
 struct grub_net_network_level_interface *
 grub_net_ipv6_get_link_local (struct grub_net_card *card,
 			      const grub_net_link_level_address_t *hwaddr)
 {
   struct grub_net_network_level_interface *inf;
-  char name[grub_strlen (card->name)
-	    + GRUB_NET_MAX_STR_HWADDR_LEN
-	    + sizeof (":link")];
+  char *name;
   char *ptr;
   grub_net_network_level_address_t addr;
 
   addr.type = GRUB_NET_NETWORK_LEVEL_PROTOCOL_IPV6;
-  addr.ipv6[0] = grub_cpu_to_be64 (0xfe80ULL << 48);
+  addr.ipv6[0] = grub_cpu_to_be64_compile_time (0xfe80ULL << 48);
   addr.ipv6[1] = grub_net_ipv6_get_id (hwaddr);
 
   FOR_NET_NETWORK_LEVEL_INTERFACES (inf)
@@ -266,6 +296,12 @@ grub_net_ipv6_get_link_local (struct grub_net_card *card,
       return inf;
   }
 
+  name = grub_malloc (grub_strlen (card->name)
+		      + GRUB_NET_MAX_STR_HWADDR_LEN
+		      + sizeof (":link"));
+  if (!name)
+    return NULL;
+
   ptr = grub_stpcpy (name, card->name);
   if (grub_net_hwaddr_cmp (&card->default_address, hwaddr) != 0)
     {
@@ -274,7 +310,7 @@ grub_net_ipv6_get_link_local (struct grub_net_card *card,
       ptr += grub_strlen (ptr);
     }
   ptr = grub_stpcpy (ptr, ":link");
-  return grub_net_add_addr (name, card, &addr, hwaddr, 0);
+  return grub_net_add_addr_real (name, card, &addr, hwaddr, 0);
 }
 
 /* FIXME: allow to specify mac address.  */
@@ -329,12 +365,14 @@ grub_cmd_ipv6_autoconf (struct grub_command *cmd __attribute__ ((unused)),
 
   for (interval = 200; interval < 10000; interval *= 2)
     {
-      /* FIXME: send router solicitation.  */
       int done = 1;
       for (j = 0; j < ncards; j++)
 	{
 	  if (slaacs[j]->slaac_counter)
 	    continue;
+	  err = grub_net_icmp6_send_router_solicit (ifaces[j]);
+	  if (err)
+	    err = GRUB_ERR_NONE;
 	  done = 0;
 	}
       if (done)
@@ -357,20 +395,6 @@ grub_cmd_ipv6_autoconf (struct grub_command *cmd __attribute__ ((unused)),
   return err;
 }
 
-static inline void
-grub_net_route_register (struct grub_net_route *route)
-{
-  grub_list_push (GRUB_AS_LIST_P (&grub_net_routes),
-		  GRUB_AS_LIST (route));
-}
-
-static inline void
-grub_net_route_unregister (struct grub_net_route *route)
-{
-  grub_list_remove (GRUB_AS_LIST (route));
-}
-
-#define FOR_NET_ROUTES(var) for (var = grub_net_routes; var; var = var->next)
 
 static int
 parse_ip (const char *val, grub_uint32_t *ip, const char **rest)
@@ -477,6 +501,8 @@ match_net (const grub_net_network_level_netaddress_t *net,
     case GRUB_NET_NETWORK_LEVEL_PROTOCOL_IPV6:
       {
 	grub_uint64_t mask[2];
+	if (net->ipv6.masksize == 0)
+	  return 1;
 	if (net->ipv6.masksize <= 64)
 	  {
 	    mask[0] = 0xffffffffffffffffULL << (64 - net->ipv6.masksize);
@@ -640,7 +666,14 @@ grub_net_route_address (grub_net_network_level_address_t addr,
 	  return GRUB_ERR_NONE;
 	}
       if (depth == 0)
-	*gateway = bestroute->gw;
+	{
+	  *gateway = bestroute->gw;
+	  if (bestroute->interface != NULL)
+	    {
+	      *interf = bestroute->interface;
+	      return GRUB_ERR_NONE;
+	    }
+	}
       curtarget = bestroute->gw;
     }
 
@@ -683,8 +716,8 @@ grub_net_addr_to_str (const grub_net_network_level_address_t *target, char *buf)
   switch (target->type)
     {
     case GRUB_NET_NETWORK_LEVEL_PROTOCOL_DHCP_RECV:
-      /* TRANSLATORS: it refers to the network address.  */
-      grub_strncpy (buf, "temporary", GRUB_NET_MAX_STR_ADDR_LEN);
+      COMPILE_TIME_ASSERT (sizeof ("temporary") < GRUB_NET_MAX_STR_ADDR_LEN);
+      grub_strcpy (buf, "temporary");
       return;
     case GRUB_NET_NETWORK_LEVEL_PROTOCOL_IPV6:
       {
@@ -882,28 +915,36 @@ grub_net_network_level_interface_register (struct grub_net_network_level_interfa
 {
   {
     char buf[GRUB_NET_MAX_STR_HWADDR_LEN];
-    char name[grub_strlen (inter->name) + sizeof ("net__mac")];
+    char *name;
     char *ptr;
     grub_net_hwaddr_to_str (&inter->hwaddress, buf);
-    grub_snprintf (name, sizeof (name), "net_%s_mac", inter->name);
+    name = grub_xasprintf ("net_%s_mac", inter->name);
+    if (!name)
+      return;
     for (ptr = name; *ptr; ptr++)
       if (*ptr == ':')
 	*ptr = '_';    
     grub_env_set (name, buf);
     grub_register_variable_hook (name, 0, hwaddr_set_env);
+    grub_env_export (name);
+    grub_free (name);
   }
 
   {
     char buf[GRUB_NET_MAX_STR_ADDR_LEN];
-    char name[grub_strlen (inter->name) + sizeof ("net__ip")];
+    char *name;
     char *ptr;
     grub_net_addr_to_str (&inter->address, buf);
-    grub_snprintf (name, sizeof (name), "net_%s_ip", inter->name);
+    name = grub_xasprintf ("net_%s_ip", inter->name);
+    if (!name)
+      return;
     for (ptr = name; *ptr; ptr++)
       if (*ptr == ':')
 	*ptr = '_';    
     grub_env_set (name, buf);
     grub_register_variable_hook (name, 0, addr_set_env);
+    grub_env_export (name);
+    grub_free (name);
   }
 
   inter->card->num_ifaces++;
@@ -914,31 +955,6 @@ grub_net_network_level_interface_register (struct grub_net_network_level_interfa
   grub_net_network_level_interfaces = inter;
 }
 
-struct grub_net_network_level_interface *
-grub_net_add_addr (const char *name, 
-		   struct grub_net_card *card,
-		   const grub_net_network_level_address_t *addr,
-		   const grub_net_link_level_address_t *hwaddress,
-		   grub_net_interface_flags_t flags)
-{
-  struct grub_net_network_level_interface *inter;
-
-  inter = grub_zalloc (sizeof (*inter));
-  if (!inter)
-    return NULL;
-
-  inter->name = grub_strdup (name);
-  grub_memcpy (&(inter->address), addr, sizeof (inter->address));
-  grub_memcpy (&(inter->hwaddress), hwaddress, sizeof (inter->hwaddress));
-  inter->flags = flags;
-  inter->card = card;
-  inter->dhcp_ack = NULL;
-  inter->dhcp_acklen = 0;
-
-  grub_net_network_level_interface_register (inter);
-
-  return inter;
-}
 
 grub_err_t
 grub_net_add_ipv4_local (struct grub_net_network_level_interface *inter,
@@ -1079,7 +1095,8 @@ grub_net_add_route (const char *name,
 grub_err_t
 grub_net_add_route_gw (const char *name,
 		       grub_net_network_level_netaddress_t target,
-		       grub_net_network_level_address_t gw)
+		       grub_net_network_level_address_t gw,
+		       struct grub_net_network_level_interface *inter)
 {
   struct grub_net_route *route;
 
@@ -1097,6 +1114,7 @@ grub_net_add_route_gw (const char *name,
   route->target = target;
   route->is_gateway = 1;
   route->gw = gw;
+  route->interface = inter;
 
   grub_net_route_register (route);
 
@@ -1122,7 +1140,7 @@ grub_cmd_addroute (struct grub_command *cmd __attribute__ ((unused)),
       err = grub_net_resolve_address (args[3], &gw);
       if (err)
 	return err;
-      return grub_net_add_route_gw (args[0], target, gw);
+      return grub_net_add_route_gw (args[0], target, gw, NULL);
     }
   else
     {
@@ -1292,25 +1310,64 @@ grub_net_open_real (const char *name)
 	    if (!ret)
 	      return NULL;
 	    ret->protocol = proto;
-	    if (server)
+	    ret->server = grub_strdup (server);
+	    if (!ret->server)
 	      {
-		ret->server = grub_strdup (server);
-		if (!ret->server)
-		  {
-		    grub_free (ret);
-		    return NULL;
-		  }
+		grub_free (ret);
+		return NULL;
 	      }
-	    else
-	      ret->server = NULL;
 	    ret->fs = &grub_net_fs;
-	    ret->offset = 0;
-	    ret->eof = 0;
 	    return ret;
 	  }
       }
       if (try == 0)
 	{
+	  const char *prefix, *root;
+	  char *prefdev, *comma;
+	  int skip = 0;
+	  grub_size_t devlen;
+
+	  /* Do not attempt to load module if it requires protocol provided
+	     by this module - it results in infinite recursion. Just continue,
+	     fail and cleanup on next iteration.
+	   */
+	  prefix = grub_env_get ("prefix");
+	  if (!prefix)
+	    continue;
+
+	  prefdev = grub_file_get_device_name (prefix);
+	  if (!prefdev)
+	    {
+	      root = grub_env_get ("root");
+	      if (!root)
+		continue;
+	      prefdev = grub_strdup (root);
+	      if (!prefdev)
+		continue;
+	    }
+
+	  if (grub_strncmp (prefdev, "pxe", sizeof ("pxe") - 1) == 0 &&
+	      (!prefdev[sizeof ("pxe") - 1] || (prefdev[sizeof("pxe") - 1] == ':')))
+	    {
+	      grub_free (prefdev);
+	      prefdev = grub_strdup ("tftp");
+	      if (!prefdev)
+		continue;
+	    }
+
+	  comma = grub_strchr (prefdev, ',');
+	  if (comma)
+	    *comma = '\0';
+	  devlen = grub_strlen (prefdev);
+
+	  if (protnamelen == devlen && grub_memcmp (prefdev, protname, devlen) == 0)
+	    skip = 1;
+
+	  grub_free (prefdev);
+
+	  if (skip)
+	    continue;
+
 	  if (sizeof ("http") - 1 == protnamelen
 	      && grub_memcmp ("http", protname, protnamelen) == 0)
 	    {
@@ -1361,7 +1418,10 @@ grub_net_fs_open (struct grub_file *file_out, const char *name)
   file->device->net->packs.last = NULL;
   file->device->net->name = grub_strdup (name);
   if (!file->device->net->name)
-    return grub_errno;
+    {
+      grub_free (file);
+      return grub_errno;
+    }
 
   err = file->device->net->protocol->open (file, name);
   if (err)
@@ -1425,7 +1485,7 @@ receive_packets (struct grub_net_card *card, int *stop_condition)
 	}
       card->opened = 1;
     }
-  while (1)
+  while (received < 100)
     {
       /* Maybe should be better have a fixed number of packets for each card
 	 and just mark them as used and not used.  */ 
@@ -1452,6 +1512,44 @@ receive_packets (struct grub_net_card *card, int *stop_condition)
   grub_print_error ();
 }
 
+static char *
+grub_env_write_readonly (struct grub_env_var *var __attribute__ ((unused)),
+			 const char *val __attribute__ ((unused)))
+{
+  return NULL;
+}
+
+grub_err_t
+grub_env_set_net_property (const char *intername, const char *suffix,
+                           const char *value, grub_size_t len)
+{
+  char *varname, *varvalue;
+  char *ptr;
+
+  varname = grub_xasprintf ("net_%s_%s", intername, suffix);
+  if (!varname)
+    return grub_errno;
+  for (ptr = varname; *ptr; ptr++)
+    if (*ptr == ':')
+      *ptr = '_';
+  varvalue = grub_malloc (len + 1);
+  if (!varvalue)
+    {
+      grub_free (varname);
+      return grub_errno;
+    }
+
+  grub_memcpy (varvalue, value, len);
+  varvalue[len] = 0;
+  grub_err_t ret = grub_env_set (varname, varvalue);
+  grub_register_variable_hook (varname, 0, grub_env_write_readonly);
+  grub_env_export (varname);
+  grub_free (varname);
+  grub_free (varvalue);
+
+  return ret;
+}
+
 void
 grub_net_poll_cards (unsigned time, int *stop_condition)
 {
@@ -1502,6 +1600,8 @@ grub_net_fs_read_real (grub_file_t file, char *buf, grub_size_t len)
 	  len -= amount;
 	  total += amount;
 	  file->device->net->offset += amount;
+	  if (grub_file_progress_hook)
+	    grub_file_progress_hook (0, 0, amount, file);
 	  if (buf)
 	    {
 	      grub_memcpy (ptr, nb->data, amount);
@@ -1528,8 +1628,9 @@ grub_net_fs_read_real (grub_file_t file, char *buf, grub_size_t len)
       if (!net->eof)
 	{
 	  try++;
-	  grub_net_poll_cards (GRUB_NET_INTERVAL, &net->stall);
-	}
+	  grub_net_poll_cards (GRUB_NET_INTERVAL +
+                               (try * GRUB_NET_INTERVAL_ADDITION), &net->stall);
+        }
       else
 	return total;
     }
@@ -1580,6 +1681,7 @@ grub_net_seek_real (struct grub_file *file, grub_off_t offset)
     file->device->net->packs.last = NULL;
     file->device->net->offset = 0;
     file->device->net->eof = 0;
+    file->device->net->stall = 0;
     err = file->device->net->protocol->open (file, file->device->net->name);
     if (err)
       return err;
@@ -1643,12 +1745,16 @@ GRUB_MOD_INIT(net)
 {
   grub_register_variable_hook ("net_default_server", defserver_get_env,
 			       defserver_set_env);
+  grub_env_export ("net_default_server");
   grub_register_variable_hook ("pxe_default_server", defserver_get_env,
 			       defserver_set_env);
+  grub_env_export ("pxe_default_server");
   grub_register_variable_hook ("net_default_ip", defip_get_env,
 			       defip_set_env);
+  grub_env_export ("net_default_ip");
   grub_register_variable_hook ("net_default_mac", defmac_get_env,
 			       defmac_set_env);
+  grub_env_export ("net_default_mac");
 
   cmd_addaddr = grub_register_command ("net_add_addr", grub_cmd_addaddr,
 					/* TRANSLATORS: HWADDRESS stands for
