@@ -49,7 +49,7 @@
 #include <grub/term.h>
 #include <grub/normal.h>
 
-#ifdef HAVE_UNIFONT_WIDTHSPEC
+#if HAVE_FONT_SOURCE
 #include "widthspec.h"
 #endif
 
@@ -387,9 +387,9 @@ grub_unicode_get_comb_type (grub_uint32_t c)
   return GRUB_UNICODE_COMB_NONE;
 }
 
-#ifdef HAVE_UNIFONT_WIDTHSPEC
+#if HAVE_FONT_SOURCE
 
-grub_ssize_t
+grub_size_t
 grub_unicode_estimate_width (const struct grub_unicode_glyph *c)
 {
   if (grub_unicode_get_comb_type (c->base))
@@ -523,11 +523,39 @@ grub_unicode_aglomerate_comb (const grub_uint32_t *in, grub_size_t inlen,
   return ptr - in;
 }
 
+static void
+revert (struct grub_unicode_glyph *visual,
+	struct grub_term_pos *pos,
+	unsigned start, unsigned end)
+{
+  struct grub_unicode_glyph t;
+  unsigned i;
+  int a = 0, b = 0;
+  if (pos)
+    {
+      a = pos[visual[start].orig_pos].x;
+      b = pos[visual[end].orig_pos].x;
+    }
+  for (i = 0; i < (end - start) / 2 + 1; i++)
+    {
+      t = visual[start + i];
+      visual[start + i] = visual[end - i];
+      visual[end - i] = t;
+
+      if (pos)
+	{
+	  pos[visual[start + i].orig_pos].x = a + b - pos[visual[start + i].orig_pos].x;
+	  pos[visual[end - i].orig_pos].x = a + b - pos[visual[end - i].orig_pos].x;
+	}
+    }
+}
+
+
 static grub_ssize_t
 bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 		struct grub_unicode_glyph *visual,
 		grub_size_t visual_len,
-		grub_ssize_t (*getcharwidth) (const struct grub_unicode_glyph *visual, void *getcharwidth_arg),
+		grub_size_t (*getcharwidth) (const struct grub_unicode_glyph *visual, void *getcharwidth_arg),
 		void *getcharwidth_arg,
 		grub_size_t maxwidth, grub_size_t startwidth,
 		grub_uint32_t contchar,
@@ -536,44 +564,34 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 {
   struct grub_unicode_glyph *outptr = visual_out;
   unsigned line_start = 0;
-  grub_ssize_t line_width = startwidth;
+  grub_ssize_t line_width;
   unsigned k;
   grub_ssize_t last_space = -1;
   grub_ssize_t last_space_width = 0;
   int lines = 0;
 
-  auto void revert (unsigned start, unsigned end);
-  void revert (unsigned start, unsigned end)
-  {
-    struct grub_unicode_glyph t;
-    unsigned i;
-    int a = 0, b = 0;
-    if (pos)
-      {
-	a = pos[visual[start].orig_pos].x;
-	b = pos[visual[end].orig_pos].x;
-      }
-    for (i = 0; i < (end - start) / 2 + 1; i++)
-      {
-	t = visual[start + i];
-	visual[start + i] = visual[end - i];
-	visual[end - i] = t;
-
-	if (pos)
-	  {
-	    pos[visual[start + i].orig_pos].x = a + b - pos[visual[start + i].orig_pos].x;
-	    pos[visual[end - i].orig_pos].x = a + b - pos[visual[end - i].orig_pos].x;
-	  }
-      }
-  }
-
   if (!visual_len)
     return 0;
 
+  if (startwidth >= maxwidth && (grub_ssize_t) maxwidth > 0)
+    {
+      if (contchar)
+	{
+	  grub_memset (outptr, 0, sizeof (visual[0]));
+	  outptr->base = contchar;
+	  outptr++;
+	}
+      grub_memset (outptr, 0, sizeof (visual[0]));
+      outptr->base = '\n';
+      outptr++;
+      startwidth = 0;
+    }
+
+  line_width = startwidth;
+
   for (k = 0; k <= visual_len; k++)
     {
       grub_ssize_t last_width = 0;
-
  
       if (pos && k != visual_len)
 	{
@@ -605,23 +623,28 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 	{	  
 	  unsigned min_odd_level = 0xffffffff;
 	  unsigned max_level = 0;
+	  unsigned kk = k;
 
 	  lines++;
 
 	  if (k != visual_len && last_space > (signed) line_start)
-	    k = last_space;
-	  else if (k != visual_len && line_start == 0 && startwidth != 0
-		   && !primitive_wrap)
 	    {
-	      k = 0;
-	      last_space_width = startwidth;
+	      kk = last_space;
+	      line_width -= last_space_width;
+	    }
+	  else if (k != visual_len && line_start == 0 && startwidth != 0
+		   && !primitive_wrap && lines == 1
+		   && line_width - startwidth < maxwidth)
+	    {
+	      kk = 0;
+	      line_width -= startwidth;
 	    }
 	  else
-	    last_space_width = line_width - last_width;
+	    line_width = last_width;
 
 	  {
 	    unsigned i;
-	    for (i = line_start; i < k; i++)
+	    for (i = line_start; i < kk; i++)
 	      {
 		if (visual[i].bidi_level > max_level)
 		  max_level = visual[i].bidi_level;
@@ -637,21 +660,21 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 	      {
 		unsigned in = line_start;
 		unsigned i;
-		for (i = line_start; i < k; i++)
+		for (i = line_start; i < kk; i++)
 		  {
 		    if (i != line_start && visual[i].bidi_level >= j
 			&& visual[i-1].bidi_level < j)
 		      in = i;
-		    if (visual[i].bidi_level >= j && (i + 1 == k
+		    if (visual[i].bidi_level >= j && (i + 1 == kk
 						 || visual[i+1].bidi_level < j))
-		      revert (in, i);
+		      revert (visual, pos, in, i);
 		  }
 	      }
 	  }
 	  
 	  {
 	    unsigned i;
-	    for (i = line_start; i < k; i++)
+	    for (i = line_start; i < kk; i++)
 	      {
 		if (is_mirrored (visual[i].base) && visual[i].bidi_level)
 		  visual[i].attributes |= GRUB_UNICODE_GLYPH_ATTRIBUTE_MIRROR;
@@ -676,7 +699,7 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 	  {
 	    int left_join = 0;
 	    unsigned i;
-	    for (i = line_start; i < k; i++)
+	    for (i = line_start; i < kk; i++)
 	      {
 		enum grub_join_type join_type = get_join_type (visual[i].base);
 		if (!(visual[i].attributes
@@ -704,7 +727,7 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 	  {
 	    int right_join = 0;
 	    signed i;
-	    for (i = k - 1; i >= 0 && (unsigned) i + 1 > line_start;
+	    for (i = kk - 1; i >= 0 && (unsigned) i + 1 > line_start;
 		 i--)
 	      {
 		enum grub_join_type join_type = get_join_type (visual[i].base);
@@ -731,9 +754,9 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 	  }		
 
 	  grub_memcpy (outptr, &visual[line_start],
-		       (k - line_start) * sizeof (visual[0]));
-	  outptr += k - line_start;
-	  if (k != visual_len)
+		       (kk - line_start) * sizeof (visual[0]));
+	  outptr += kk - line_start;
+	  if (kk != visual_len)
 	    {
 	      if (contchar)
 		{
@@ -746,15 +769,14 @@ bidi_line_wrap (struct grub_unicode_glyph *visual_out,
 	      outptr++;
 	    }
 
-	  if ((signed) k == last_space)
-	    k++;
+	  if ((signed) kk == last_space)
+	    kk++;
 
-	  line_start = k;
-	  line_width -= last_space_width;
-	  if (pos && k != visual_len)
+	  line_start = kk;
+	  if (pos && kk != visual_len)
 	    {
-	      pos[visual[k].orig_pos].x = 0;
-	      pos[visual[k].orig_pos].y = lines;
+	      pos[visual[kk].orig_pos].x = 0;
+	      pos[visual[kk].orig_pos].y = lines;
 	    }
 	}
     }
@@ -767,7 +789,7 @@ static grub_ssize_t
 grub_bidi_line_logical_to_visual (const grub_uint32_t *logical,
 				  grub_size_t logical_len,
 				  struct grub_unicode_glyph *visual_out,
-				  grub_ssize_t (*getcharwidth) (const struct grub_unicode_glyph *visual, void *getcharwidth_arg),
+				  grub_size_t (*getcharwidth) (const struct grub_unicode_glyph *visual, void *getcharwidth_arg),
 				  void *getcharwidth_arg,
 				  grub_size_t maxwidth, grub_size_t startwidth,
 				  grub_uint32_t contchar,
@@ -790,34 +812,34 @@ grub_bidi_line_logical_to_visual (const grub_uint32_t *logical,
   unsigned cur_level;
   int bidi_needed = 0;
 
-  auto void push_stack (unsigned new_override, unsigned new_level);
-  void push_stack (unsigned new_override, unsigned new_level)
-  {
-    if (new_level > GRUB_BIDI_MAX_EXPLICIT_LEVEL)
-      {
-	invalid_pushes++;
-	return;
-      }
-    stack_level[stack_depth] = cur_level;
-    stack_override[stack_depth] = cur_override;
-    stack_depth++;
-    cur_level = new_level;
-    cur_override = new_override;
+#define push_stack(new_override, new_level)		\
+  {							\
+    if (new_level > GRUB_BIDI_MAX_EXPLICIT_LEVEL)	\
+      {							\
+      invalid_pushes++;					\
+      }							\
+    else						\
+      {							\
+      stack_level[stack_depth] = cur_level;		\
+      stack_override[stack_depth] = cur_override;	\
+      stack_depth++;					\
+      cur_level = new_level;				\
+      cur_override = new_override;			\
+      }							\
   }
 
-  auto void pop_stack (void);
-  void pop_stack (void)
-  {
-    if (invalid_pushes)
-      {
-	invalid_pushes--;
-	return;
-      }
-    if (!stack_depth)
-      return;
-    stack_depth--;
-    cur_level = stack_level[stack_depth];
-    cur_override = stack_override[stack_depth];
+#define pop_stack()				\
+  {						\
+    if (invalid_pushes)				\
+      {						\
+      invalid_pushes--;				\
+      }						\
+    else if (stack_depth)			\
+      {						\
+      stack_depth--;				\
+      cur_level = stack_level[stack_depth];	\
+      cur_override = stack_override[stack_depth];	\
+      }							\
   }
 
   visual = grub_malloc (sizeof (visual[0]) * logical_len);
@@ -907,6 +929,7 @@ grub_bidi_line_logical_to_visual (const grub_uint32_t *logical,
 	  case GRUB_BIDI_TYPE_R:
 	  case GRUB_BIDI_TYPE_AL:
 	    bidi_needed = 1;
+	    /* Fallthrough.  */
 	  default:
 	    {
 	      if (join_state == JOIN_FORCE)
@@ -1121,11 +1144,21 @@ grub_bidi_line_logical_to_visual (const grub_uint32_t *logical,
   }
 }
 
+static int
+is_visible (const struct grub_unicode_glyph *gl)
+{
+  if (gl->ncomb)
+    return 1;
+  if (gl->base == GRUB_UNICODE_LRM || gl->base == GRUB_UNICODE_RLM)
+    return 0;
+  return 1;
+}
+
 grub_ssize_t
 grub_bidi_logical_to_visual (const grub_uint32_t *logical,
 			     grub_size_t logical_len,
 			     struct grub_unicode_glyph **visual_out,
-			     grub_ssize_t (*getcharwidth) (const struct grub_unicode_glyph *visual, void *getcharwidth_arg),
+			     grub_size_t (*getcharwidth) (const struct grub_unicode_glyph *visual, void *getcharwidth_arg),
 			     void *getcharwidth_arg,
 			     grub_size_t max_length, grub_size_t startwidth,
 			     grub_uint32_t contchar, struct grub_term_pos *pos, int primitive_wrap)
@@ -1133,7 +1166,7 @@ grub_bidi_logical_to_visual (const grub_uint32_t *logical,
   const grub_uint32_t *line_start = logical, *ptr;
   struct grub_unicode_glyph *visual_ptr;
   *visual_out = visual_ptr = grub_malloc (3 * sizeof (visual_ptr[0])
-					  * logical_len);
+					  * (logical_len + 2));
   if (!visual_ptr)
     return -1;
   for (ptr = logical; ptr <= logical + logical_len; ptr++)
@@ -1141,6 +1174,7 @@ grub_bidi_logical_to_visual (const grub_uint32_t *logical,
       if (ptr == logical + logical_len || *ptr == '\n')
 	{
 	  grub_ssize_t ret;
+	  grub_ssize_t i, j;
 	  ret = grub_bidi_line_logical_to_visual (line_start,
 						  ptr - line_start,
 						  visual_ptr,
@@ -1159,7 +1193,10 @@ grub_bidi_logical_to_visual (const grub_uint32_t *logical,
 	      grub_free (*visual_out);
 	      return ret;
 	    }
-	  visual_ptr += ret; 
+	  for (i = 0, j = 0; i < ret; i++)
+	    if (is_visible(&visual_ptr[i]))
+	      visual_ptr[j++] = visual_ptr[i];
+	  visual_ptr += j;
 	  line_start = ptr;
 	  if (ptr != logical + logical_len)
 	    {
