@@ -230,38 +230,37 @@ putstr (struct grub_term_output *term, const char *str)
     data->put (term, *str++);
 }
 
-grub_uint16_t
+struct grub_term_coordinate
 grub_terminfo_getxy (struct grub_term_output *term)
 {
   struct grub_terminfo_output_state *data
     = (struct grub_terminfo_output_state *) term->data;
 
-  return ((data->xpos << 8) | data->ypos);
+  return data->pos;
 }
 
 void
 grub_terminfo_gotoxy (struct grub_term_output *term,
-		      grub_uint8_t x, grub_uint8_t y)
+		      struct grub_term_coordinate pos)
 {
   struct grub_terminfo_output_state *data
     = (struct grub_terminfo_output_state *) term->data;
 
-  if (x > grub_term_width (term) || y > grub_term_height (term))
+  if (pos.x > grub_term_width (term) || pos.y > grub_term_height (term))
     {
-      grub_error (GRUB_ERR_BUG, "invalid point (%u,%u)", x, y);
+      grub_error (GRUB_ERR_BUG, "invalid point (%u,%u)", pos.x, pos.y);
       return;
     }
 
   if (data->gotoxy)
-    putstr (term, grub_terminfo_tparm (data->gotoxy, y, x));
+    putstr (term, grub_terminfo_tparm (data->gotoxy, pos.y, pos.x));
   else
     {
-      if ((y == data->ypos) && (x == data->xpos - 1))
+      if ((pos.y == data->pos.y) && (pos.x == data->pos.x - 1))
 	data->put (term, '\b');
     }
 
-  data->xpos = x;
-  data->ypos = y;
+  data->pos = pos;
 }
 
 /* Clear the screen.  */
@@ -272,7 +271,7 @@ grub_terminfo_cls (struct grub_term_output *term)
     = (struct grub_terminfo_output_state *) term->data;
 
   putstr (term, grub_terminfo_tparm (data->cls));
-  grub_terminfo_gotoxy (term, 0, 0);
+  grub_terminfo_gotoxy (term, (struct grub_term_coordinate) { 0, 0 });
 }
 
 void
@@ -360,42 +359,42 @@ grub_terminfo_putchar (struct grub_term_output *term,
 
     case '\b':
     case 127:
-      if (data->xpos > 0)
-	data->xpos--;
+      if (data->pos.x > 0)
+	data->pos.x--;
     break;
 
     case '\n':
-      if (data->ypos < grub_term_height (term) - 1)
-	data->ypos++;
+      if (data->pos.y < grub_term_height (term) - 1)
+	data->pos.y++;
       break;
 
     case '\r':
-      data->xpos = 0;
+      data->pos.x = 0;
       break;
 
     default:
-      if (data->xpos + c->estimated_width >= grub_term_width (term) + 1)
+      if ((int) data->pos.x + c->estimated_width >= (int) grub_term_width (term) + 1)
 	{
-	  data->xpos = 0;
-	  if (data->ypos < grub_term_height (term) - 1)
-	    data->ypos++;
+	  data->pos.x = 0;
+	  if (data->pos.y < grub_term_height (term) - 1)
+	    data->pos.y++;
 	  data->put (term, '\r');
 	  data->put (term, '\n');
 	}
-      data->xpos += c->estimated_width;
+      data->pos.x += c->estimated_width;
       break;
     }
 
   data->put (term, c->base);
 }
 
-grub_uint16_t
+struct grub_term_coordinate
 grub_terminfo_getwh (struct grub_term_output *term)
 {
   struct grub_terminfo_output_state *data
     = (struct grub_terminfo_output_state *) term->data;
 
-  return (data->width << 8) | data->height;
+  return data->size;
 }
 
 static void
@@ -461,32 +460,50 @@ grub_terminfo_readkey (struct grub_term_input *term, int *keys, int *len,
 	{'@', GRUB_TERM_KEY_INSERT},
       };
 
-    static struct
-    {
-      char key;
-      unsigned ascii;
-    }
-    four_code_table[] =
+    static unsigned four_code_table[] =
       {
-	{'1', GRUB_TERM_KEY_HOME},
-	{'3', GRUB_TERM_KEY_DC},
-	{'5', GRUB_TERM_KEY_PPAGE},
-	{'6', GRUB_TERM_KEY_NPAGE}
+	[1] = GRUB_TERM_KEY_HOME,
+	[3] = GRUB_TERM_KEY_DC,
+	[5] = GRUB_TERM_KEY_PPAGE,
+	[6] = GRUB_TERM_KEY_NPAGE,
+	[7] = GRUB_TERM_KEY_HOME,
+	[8] = GRUB_TERM_KEY_END,
+	[17] = GRUB_TERM_KEY_F6,
+	[18] = GRUB_TERM_KEY_F7,
+	[19] = GRUB_TERM_KEY_F8,
+	[20] = GRUB_TERM_KEY_F9,
+	[21] = GRUB_TERM_KEY_F10,
+	[23] = GRUB_TERM_KEY_F11,
+	[24] = GRUB_TERM_KEY_F12,
       };
     char fx_key[] = 
       { 'P', 'Q', 'w', 'x', 't', 'u',
-        'q', 'r', 'p', 'M', 'A', 'B' };
+        'q', 'r', 'p', 'M', 'A', 'B', 'H', 'F' };
     unsigned fx_code[] = 
 	{ GRUB_TERM_KEY_F1, GRUB_TERM_KEY_F2, GRUB_TERM_KEY_F3,
 	  GRUB_TERM_KEY_F4, GRUB_TERM_KEY_F5, GRUB_TERM_KEY_F6,
 	  GRUB_TERM_KEY_F7, GRUB_TERM_KEY_F8, GRUB_TERM_KEY_F9,
-	  GRUB_TERM_KEY_F10, GRUB_TERM_KEY_F11, GRUB_TERM_KEY_F12 };
+	  GRUB_TERM_KEY_F10, GRUB_TERM_KEY_F11, GRUB_TERM_KEY_F12,
+	  GRUB_TERM_KEY_HOME, GRUB_TERM_KEY_END };
     unsigned i;
 
     if (c == '\e')
       {
 	CONTINUE_READ;
 
+	if (c == 'O')
+	  {
+	    CONTINUE_READ;
+
+	    for (i = 0; i < ARRAY_SIZE (fx_key); i++)
+	      if (fx_key[i] == c)
+		{
+		  keys[0] = fx_code[i];
+		  *len = 1;
+		  return;
+		}
+	  }
+
 	if (c != '[')
 	  return;
       }
@@ -503,6 +520,15 @@ grub_terminfo_readkey (struct grub_term_input *term, int *keys, int *len,
 
     switch (c)
       {
+      case '[':
+	CONTINUE_READ;
+	if (c >= 'A' && c <= 'E')
+	  {
+	    keys[0] = GRUB_TERM_KEY_F1 + c - 'A';
+	    *len = 1;
+	    return;
+	  }
+	return;
       case 'O':
 	CONTINUE_READ;
 	for (i = 0; i < ARRAY_SIZE (fx_key); i++)
@@ -535,18 +561,26 @@ grub_terminfo_readkey (struct grub_term_input *term, int *keys, int *len,
 	  return;
 	}	  
 
-      default:
-	for (i = 0; i < ARRAY_SIZE (four_code_table); i++)
-	  if (four_code_table[i].key == c)
+      case '1' ... '9':
+	{
+	  unsigned val = c - '0';
+	  CONTINUE_READ;
+	  if (c >= '0' && c <= '9')
 	    {
+	      val = val * 10 + (c - '0');
 	      CONTINUE_READ;
-	      if (c != '~')
-		return;
-	      keys[0] = three_code_table[i].ascii;
-	      *len = 1;
-	      return;
 	    }
-	return;
+	  if (c != '~')
+	    return;
+	  if (val >= ARRAY_SIZE (four_code_table)
+	      || four_code_table[val] == 0)
+	    return;
+	  keys[0] = four_code_table[val];
+	  *len = 1;
+	  return;
+	}
+	default:
+	  return;
       }
   }
 #undef CONTINUE_READ
@@ -654,8 +688,8 @@ print_terminfo (void)
 		 grub_terminfo_get_current(cur),
 		 encoding_names[(cur->flags & GRUB_TERM_CODE_TYPE_MASK)
 				>> GRUB_TERM_CODE_TYPE_SHIFT],
-		 ((struct grub_terminfo_output_state *) cur->data)->width,
-	         ((struct grub_terminfo_output_state *) cur->data)->height);
+		 ((struct grub_terminfo_output_state *) cur->data)->pos.x,
+	         ((struct grub_terminfo_output_state *) cur->data)->pos.y);
 
   return GRUB_ERR_NONE;
 }
@@ -728,8 +762,8 @@ grub_cmd_terminfo (grub_extcmd_context_t ctxt, int argc, char **args)
 	  {
 	    struct grub_terminfo_output_state *data
 	      = (struct grub_terminfo_output_state *) cur->data;
-	    data->width = w;
-	    data->height = h;
+	    data->size.x = w;
+	    data->size.y = h;
 	  }
 
 	if (argc == 1)
@@ -745,11 +779,7 @@ grub_cmd_terminfo (grub_extcmd_context_t ctxt, int argc, char **args)
 
 static grub_extcmd_t cmd;
 
-#if defined (GRUB_MACHINE_IEEE1275) || defined (GRUB_MACHINE_MIPS_LOONGSON) || defined (GRUB_MACHINE_MIPS_QEMU_MIPS) || defined (GRUB_MACHINE_ARC)
-void grub_terminfo_init (void)
-#else
 GRUB_MOD_INIT(terminfo)
-#endif
 {
   cmd = grub_register_extcmd ("terminfo", grub_cmd_terminfo, 0,
 			      N_("[[-a|-u|-v] [-g WxH] TERM [TYPE]]"),
@@ -757,11 +787,7 @@ GRUB_MOD_INIT(terminfo)
 			      options);
 }
 
-#if defined (GRUB_MACHINE_IEEE1275) || defined (GRUB_MACHINE_MIPS_LOONGSON) || defined (GRUB_MACHINE_MIPS_QEMU_MIPS) || defined (GRUB_MACHINE_ARC)
-void grub_terminfo_fini (void)
-#else
 GRUB_MOD_FINI(terminfo)
-#endif
 {
   grub_unregister_extcmd (cmd);
 }
