@@ -54,6 +54,7 @@
 
 /* The offset of GRUB_TOTAL_MODULE_SIZE.  */
 #define GRUB_KERNEL_SPARC64_IEEE1275_TOTAL_MODULE_SIZE	0x8
+#define GRUB_KERNEL_ARM_STACK_SIZE 0x40000
 
 #define GRUB_BOOT_SPARC64_IEEE1275_LIST_SIZE	12
 
@@ -99,14 +100,20 @@
 #define GRUB_KERNEL_I386_COREBOOT_MOD_ALIGN	0x1
 #define GRUB_KERNEL_I386_MULTIBOOT_MOD_ALIGN	GRUB_KERNEL_I386_COREBOOT_MOD_ALIGN
 
+#define GRUB_KERNEL_X86_64_XEN_MOD_ALIGN	0x8
+#define GRUB_KERNEL_I386_XEN_MOD_ALIGN	0x8
+
 /* Non-zero value is only needed for PowerMacs.  */
+#define GRUB_KERNEL_X86_64_XEN_MOD_GAP 0x0
+#define GRUB_KERNEL_I386_XEN_MOD_GAP 0x0
 #define GRUB_KERNEL_I386_IEEE1275_MOD_GAP 0x0
 #define GRUB_KERNEL_I386_COREBOOT_MOD_GAP 0x0
 #define GRUB_KERNEL_SPARC64_IEEE1275_MOD_GAP 0x0
 #define GRUB_KERNEL_ARM_UBOOT_MOD_GAP 0x0
 
 #define GRUB_KERNEL_POWERPC_IEEE1275_MOD_ALIGN 0x1000
-#define GRUB_KERNEL_SPARC64_IEEE1275_MOD_ALIGN 0x1
+#define GRUB_KERNEL_SPARC64_IEEE1275_LOG_MOD_ALIGN 3
+#define GRUB_KERNEL_SPARC64_IEEE1275_MOD_ALIGN (1 << GRUB_KERNEL_SPARC64_IEEE1275_LOG_MOD_ALIGN)
 
 #define GRUB_KERNEL_MIPS_LOONGSON_MOD_ALIGN 0x1
 #define GRUB_KERNEL_MIPS_ARC_MOD_ALIGN 0x1
@@ -114,7 +121,6 @@
 
 #define GRUB_KERNEL_ARM_UBOOT_MOD_ALIGN 	0x8
 #define GRUB_KERNEL_ARM_UBOOT_TOTAL_MODULE_SIZE	0x4
-#define GRUB_KERNEL_ARM_UBOOT_LINK_ADDR		0x08000000
 
 /* Minimal gap between _end and the start of the modules.  It's a hack
    for PowerMac to prevent "CLAIM failed" error.  The real fix is to
@@ -147,7 +153,7 @@ struct grub_pc_bios_boot_blocklist
   grub_uint64_t start;
   grub_uint16_t len;
   grub_uint16_t segment;
-} __attribute__ ((packed));
+} GRUB_PACKED;
 #endif
 
 #endif
