@@ -290,16 +290,16 @@ grub_script_arg_add (struct grub_parser_param *state,
 		     grub_script_arg_type_t type, char *str);
 
 struct grub_script *grub_script_parse (char *script,
-				       grub_reader_getline_t getline,
-				       void *getline_data);
+				       grub_reader_getline_t getline_func,
+				       void *getline_func_data);
 void grub_script_free (struct grub_script *script);
 struct grub_script *grub_script_create (struct grub_script_cmd *cmd,
 					struct grub_script_mem *mem);
 
 struct grub_lexer_param *grub_script_lexer_init (struct grub_parser_param *parser,
 						 char *script,
-						 grub_reader_getline_t getline,
-						 void *getline_data);
+						 grub_reader_getline_t getline_func,
+						 void *getline_func_data);
 void grub_script_lexer_fini (struct grub_lexer_param *);
 void grub_script_lexer_ref (struct grub_lexer_param *);
 void grub_script_lexer_deref (struct grub_lexer_param *);
@@ -387,7 +387,8 @@ grub_script_execute_arglist_to_argv (struct grub_script_arglist *arglist, int *c
 
 grub_err_t
 grub_normal_parse_line (char *line,
-			grub_reader_getline_t getline, void *getline_data);
+			grub_reader_getline_t getline_func,
+			void *getline_func_data);
 
 static inline struct grub_script *
 grub_script_ref (struct grub_script *script)
